/*****************************************************************************
** FILE NAME:       main.c
**
** DESCRIPTION:     Help File Builder Utility
**
**                  Main Module
**
**  This program generates binary help files readable by programs that use
**  the Window Manager help facility.  It reads a help source text file
**  and generates a binary help file and C and assembly language include
**  files which define symbols used to refer to help topics.
**
**  Command line syntax
**  -------------------
**
**      helpbld <srcfile> [hlpfile] [incfile]
**
**  <srcfile>           The name of the help text source file.  The format
**                      of this file is described below.  A .txt extension
**                      is assumed if none is specified.
**
**  [hlpfile]           Optional destination help file name.  If no extension
**                      is specified, ".hlp" is appended.  If nothing is
**                      specified, the <srcfile> name is used with a ".hlp"
**                      extension.
**
**  [incfile]           Optional destination include file name (no extension).
**                      ".h" is appended to the include file.  If nothing is
**                      specified, the <srcfile> name is used.
**
**
**  Help Text Source File Format
**  ----------------------------
**
**  The help source text consists of a series of topic names and associated
**  explanatory text.  There are two types of topics: help topics and
**  glossary topics.
**
**  A help topic contains a thorough discussion of a particular subject.  Its
**  text is displayed in the help viewer window and can be any size.  A help
**  topic can be associated with a menu item, a dialog box, or with a word or
**  phrase in another help topic.  The user can point to a menu item and
**  press the help key to view the associated help topic; You can
**  press the help key while a dialog box is displayed to learn about that
**  dialog box;  Or, while viewing a help topic, the user can point to a
**  hilighted section of help text which is linked to another help topic.
**
**  A glossary topic merely defines the meaning of a word or phrase found
**  in the text of a help topic.  When the user points to a word or phrase
**  of help text that has an associated glossary topic, a window pops up
**  showing the glossary text.  Since the window which displays glossary
**  entries is limited in size and can't be scrolled, a glossary topic can't
**  exceed about 20 lines (by about 75 columns) of text.  (helpbld displays
**  a warning message if the limit is exceeded.)
**
**  The only assumption made about the ordering of topics in the source
**  file is that the very first topic should be the main index or table
**  of contents -- a set of links to all the major topics in the help
**  file.  When the "Index" button in the help viewer dialog is pressed,
**  this first topic is always displayed.  It is also best to have related
**  topics arranged sequentially in the help file, since this is the order
**  in which topics are presented when the user presses the "Next" button
**  to view the next topic.
**
**  Each topic in the source file has this format:
**
**          <symbolic name> [flag flag ...]
**          <title text>
**          <explanatory text>
**
**  Blank lines before the symbolic name and after the explanatory text are
**  ignored, as are lines beginning with a pound-sign (#).
**
**  <symbolic name>     This is the symbolic constant to be used in program
**                      code and data and in the help source text itself to
**                      refer to this topic (e.g. to relate a menu item or
**                      phrase of help text to this help topic).  It is
**                      assigned a numeric value using the "#define"
**                      directive in the include file generated by helpbld.
**
**                      Symbolic names are case sensitive and are significant
**                      only up to 31 characters in length within helpbld.
**
**  [flag flag ...]     Optional topic flags:
**
**                          HELP        Indicates this is a help topic
**                          GLOS        Indicates this is a glossary topic
**
**                      These flags control the positioning of the help
**                      viewer window and are meaningful only for help
**                      topics, not glossary topics:
**
**                          LEFT        Left justify window
**                          HCENTER     Center window horizontally
**                          RIGHT       Right justify window
**
**                          TOP         Top justify window
**                          VCENTER     Center window vertically
**                          BOTTOM      Bottom justify window
**
**                      Flags may be in upper or lower case.
**
**                      The default flags are HELP HCENTER VCENTER.
**
**  <title text>        This is the topic title that appears at the top
**                      of the help viewer window (help topics) or the
**                      glossary window (glossary topics) when the
**                      associated text is being viewed.  It may be
**                      only one line long and is limited in size by the
**                      width of the window.  If the maximum size is
**                      exceeded, helpbld displays a warning message.
**
**  <explanatory text>  This is the help text which illuminates the topic
**                      of interest.  It may be more than one line long and
**                      paragraphs are reformatted by helpbld to fit properly
**                      in the appropriate window.  Line breaks are ignored
**                      (actually, treated as space characters) unless the
**                      line is terminated by a backslash (see below).
**                      Special escape sequences may be embedded in
**                      the text:
**
**                      \       A backslash at the end of a line marks the
**                              end of a paragraph, i.e., forces a new line
**                              in the viewer window.
**
**                              A blank line is considered equivalent to a
**                              line containing only a backslash.  In both
**                              cases, a blank line is inserted in the help
**                              viewer window.
**
**                      \e      End of text.  Marks the end of the explanatory
**                              text for a topic.  Text must always be
**                              terminated by \e.
**
**                      \c      Hilight character.  Hilights the next
**                              character only.
**
**                      \s      Hilight string.  Hilights all subsequent
**                              characters until another \s is encountered.
**
**                              (Note: The \c and \s sequences simply
**                              hilight text, they do not establish links
**                              to other topics.)
**
**                      \l      Topic link.  Associates a string of text
**                              with another topic, either a glossary topic
**                              or help topic, which is displayed when
**                              the user points to the text with the mouse
**                              or tabs to it and presses the Enter key.
**                              Topic links have this format:
**
**                                  \l<symbol>:<text>\l
**
**                              <symbol> is the symbolic name of the topic
**                              to link to, and <text>, which may span
**                              multiple lines, is the help text to link
**                              to that topic.  The <symbol> and <text> must
**                              be separated by a colon (:) as shown.
**
**                              The \l sequence is only valid in a help topic,
**                              not a glossary topic.
**
**                              Example:
**
**                                  And the Lord Jesus H. cast out the
**                                  \lGLOS_TUSSOCK:tussocks\l until no
**                                  golf courses remained in Jerusalem.
**                                  He said unto \lHELP_TREVINO:Trevino\l,
**                                  "The birdies shall bogie on the eagles."
**
**                      \\      A backslash character.
**
**                      The character following the backslash in an escape
**                      sequence may be in upper or lower case.
**
**  Example Help File:
**
**      # These are all comments
**      # in a sample help file
**      # This means you
**
**      HELP_DOOMSDAY right bottom
**      The Doomsday Machine
**      Under the authority granted me as director of weapons research
**      and development, I commissioned a study of such a device by the
**      \lHELP_BLAND:Bland Corporation\l.  Based on the findings of the
**      report, my conclusion was that this was not a practical deterrent
**      for reasons which, at this moment, must be all too obvious.
**      \e
**
**      HELP_BLAND top
**      The Bland Corporation
**      Located in the heart of the western center of our National Security
**      infrastructure, the Bland Corporation is a vital cog in our
**      effort to stem communist aggression in its many forms around
**      the world.
**
**      It was founded in 1946 by Branford Bland, the inventor
**      of the \lGLOS_FLUORIDE:Fluoride Detector\l.
**      \e
**
**      GLOS_FLUORIDE GLOS
**      Fluoride Detector
**      A device used to detect the presence and concentration of
**      fluoridation in water, the most monstrously conceived and
**      dangerous communist plot we've ever had to face.
**      \e
**
**      HELP_SONG LEFT TOP
**      A Song
**      Ohhhh theeerrreee aint no place\
**      like a hole in the ground,\
**      a hole in the ground,\
**      a hole in the ground.\
**
**      There aint no place\
**      like a hole in the ground,\
**      with a big fat goon\
**      a floatin' around.\
**      \e
**
**  Note the location of the paragraph markers (backslashes at the end of
**  some lines) and other escape sequences in the help text for each topic.
**  In the first topic, the two paragraphs of help text would be reformatted
**  as necessary to fit in the help viewer window.  In the last topic, each
**  line is terminated by a paragraph marker and would appear exactly as
**  shown in the help viewer window (unless the window were unreasonably
**  narrow).
**
**
**  Program Exit Codes
**  ------------------
**
**      0       Everything went fine.
**  not 0       Everything messed up.
**
**
**  Binary Help File Format
**  -----------------------
**
**  The binary help file generated by this program and read by the Window
**  Manager help facility is divided into three sections:
**
**      +--------------------+
**      |       header       |
**      +--------------------+
**      |  topic descriptor  |
**      |       array        |
**      +--------------------+
**      |                    |
**      |     topic text     |
**      |                    |
**      +--------------------+
**
**  The header (type HELP_HDR) is fixed size and contains a doubleword
**  signature used to validate the help file, and the number of topics
**  contained in the help file.
**
**  The topic descriptor array contains one entry for each topic in the
**  help file (type HELP_TOPIC).  Each topic descriptor defines the
**  location of the text for a topic in the Topic Text section, the
**  number of lines of text, and the topic flags as specified in the
**  symbol/flags line of the help source file.
**
**  The topic text section contains all the titles and explanatory text.
**  It is an array of variable length strings, terminated by zeros,
**  one for each line of formatted text.  The first line of text for each
**  topic is the title line, which cannot contain any escape sequences.
**  All subsequent lines are explanatory text, which may contain escape
**  sequences (starting with WIN_ESC_CHAR bytes) to highlight text and
**  define links to other topics.
**
**  A special escape sequence is used to indicate a topic link.  It precedes
**  the actual text which is linked to the specified topic and is formatted
**  as:
**
**          db      "This is any text preceding the link"
**          db      WIN_ESC_CHAR, WIN_ESC_LINK              ; the link itself
**          dw      encoded topic index
**          db      EASx, "This is the text to link", EAS0
**          db      "This is any text following the link"
**          db      0                                       ; end of line
**
**  The encoded topic index is a binary word value (C type unsigned) which
**  is specially encoded so that neither byte is a zero or WIN_ESC_CHAR
**  (0xff).  This is done to simplify the code in the Window Manager help
**  viewer that must scan forward AND BACKWARD through the help text looking
**  for the ends of lines and for escape sequences.  The macros that encode
**  and decode these topic indexes are called, mystifyingly enough,
**  WIN_HELP_LINK_ENCODE and WIN_HELP_LINK_DECODE.
**
******************************************************************************
**
** static void     Title (void);
** static void     Parse_Command (int argc, char **argv);
** static void     Open_Files (void);
** void     Rewind_Source (void);
** static void     Close_Files (void);
** void     main (int argc, char **argv);
**
******************************************************************************
** REVISION HISTORY:
**   Wed May 31, 1989 - initiated by Ken Broomfield
**
**  8/14/89     KWB     Added hypertext linkage capability.
**
*****************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "hlpmgrif.h"
#include "hlpbld.h"
#include "hlppass1.h"
#include "hlppass2.h"
#include "hlpsymbol.h"

/******************************************************************************
**          FORWARD FUNCTION DECLARATIONS (GENERATED BY MM.EXE)
******************************************************************************/
static void     Title (void);
static void     Parse_Command (int argc, char *argv[]);
static void     Open_Files (void);
void     Rewind_Source (void);
static void     Close_Files (void);
int     main (int argc, char *argv[]);
/******************************************************************************
**        END OF FORWARD FUNCTION DECLARATIONS (GENERATED BY MM.EXE)
******************************************************************************/


/*****************************************************************************
**                               Global Data
*****************************************************************************/

/*
**  Files and filenames
*/

char Source_Name[MAX_PATHLEN];    /* source name */
char Help_Name[MAX_PATHLEN];      /* destination name */
char Include_Name[MAX_PATHLEN];   /* include name w/o extension */

FILE *Source_File;           /* source text file */
FILE *Help_File;             /* destination help file */
FILE *Include_File;          /* destination include file */

/*
**  Help file variables
*/

int   Topic_Count;            /* number of topics in file */
int   Line_Number = 0;        /* current line number */
int   Last_Eol = TRUE;        /* last Get_Source_Char was '\n' */

/*
**  Error messages
*/

char Source_File_Error[] = "\nError reading source file\n";
char Help_File_Error[]   = "\nError writing help file\n";
char Inc_File_Error[]    = "\nError writing include file\n";


/*
**	Misc other
*/

int  verbose = FALSE;		/* verbose printing flag */
char *Pgm_Name;				/* name of program */

/*.
******************************************************************************
**
**  function -  Title:
**
**  desc -      Display the opening title
**
******************************************************************************
*/

static void     Title (void)

{
#if 0
	printf ("+---------------------------------+\n");
	printf ("|                                 |\n");
	printf ("|             helpbld             |\n");
	printf ("|        Help File Builder        |\n");
	printf ("|                                 |\n");
	printf ("+---------------------------------+\n");
#endif
} /* Title */


/*.
******************************************************************************
**
**  function -  Usage
**
**  desc -      Print a usage message
**
**  entry -     NONE
**
**  exit -      NONE
**
**  notes -
**
******************************************************************************
*/

static void     Usage (void)

{
	printf("Syntax: %s [-v] <srcfile> [hlpfile] [incfile]\n", Pgm_Name);
	printf("\n");
	printf("Options:\n");
	printf("-v         Verbose flag\n");
	printf("\n");
	printf("<srcfile>  The name of the help text source file.  The format\n");
	printf("           of this file is described below.  A .txt extension\n");
	printf("           is assumed if none is specified.\n");
	printf("\n");
	printf("[hlpfile]  Optional dest help file name.  If no extension\n");
	printf("           is specified, '.hlp' is appended.  If nothing is\n");
	printf("           specified, the <srcfile> name is used with a '.hlp'\n");
	printf("           extension.\n");
	printf("\n");
	printf("[incfile]  Optional dest include file name (no extension).\n");
	printf("           '.h' is appended to the include file. If nothing is\n");
	printf("           specified, the <srcfile> name is used.\n");
}

/*.
******************************************************************************
**
**  function -  Parse_Command:
**
**  desc -      Parse the command line
**
**  entry -     argc, argv          command line
**
**  exit -      globals:
**              Source_Name
**              Help_Name
**              Include_Name
**
**  notes -
**
******************************************************************************
*/

static void     Parse_Command (int argc, char *argv[])

{
	int arg_ptr = 1;
	char *dotptr;

	if (argc <= 1)
	{
		Usage();
		exit(1);
		/*NOTREACHED*/
	}

	if (strcmp(argv[arg_ptr], "-?") == 0)
	{
		Usage();
		exit(0);
		/*NOTREACHED*/
	}

	if (strcmp(argv[arg_ptr], "-v") == 0)
	{
		verbose = TRUE;
		arg_ptr++;
	}

	/*
    **  If no arguments specified, display help
    */

	if (arg_ptr >= argc)
	{
		Usage();
		exit(1);
		/*NOTREACHED*/
	}

	/*
    **  Copy the source filename and append an extension if necessary
    */

	strcpy (Source_Name, argv[arg_ptr++]);
	dotptr = strrchr (Source_Name, '.');
	if (dotptr == (char *)NULL  ||  strrchr (Source_Name, '/') > dotptr)
		strcat (Source_Name, ".txt");

	/*
    **  If there's a help file name, use it; else use the base source name
    */

	if (arg_ptr < argc)
		strcpy (Help_Name, argv[arg_ptr++]);
		else
	{
		strcpy (Help_Name, Source_Name);
		*strrchr (Help_Name, '.') = 0;
	}
	dotptr = strrchr (Help_Name, '.');
	if (dotptr == (char *)NULL  ||  strrchr (Help_Name, '/') > dotptr)
		strcat (Help_Name, ".hlp");

	/*
    **  If there's an include file name, use it and remove any extension;
    **  else use the base source name
    */

	if (arg_ptr < argc)
	{
		strcpy (Include_Name, argv[arg_ptr++]);
		dotptr = strrchr (Include_Name, '.');
		if (dotptr != (char *)NULL  &&  strrchr (Include_Name, '/') < dotptr)
			*dotptr = 0;
	}
	else
	{
		strcpy (Include_Name, Source_Name);
		*strrchr (Include_Name, '.') = 0;
	}

	/*
    **  Tell them which files will be operated on
    */

	if (verbose)
	{
		printf ("Reading %s\n", Source_Name);
		printf ("Writing %s\n", Help_Name);
		printf ("    and %s.h\n", Include_Name);
	}
} /* Parse_Command */


/*.
******************************************************************************
**
**  function -  Open_Files:
**
**  desc -      Open all input and output files processed by this program
**
**  entry -     globals:
**              Source_Name
**              Help_Name
**              Include_Name
**
**  exit -
**
**  notes -     If an error occurs, a message is displayed and the program
**              terminates.
**
******************************************************************************
*/

static void     Open_Files (void)

{
	char    namebuf [MAX_PATHLEN];

	/*
    **  Open the source help text file
    */

	Source_File = fopen (Source_Name, "r");
	if (Source_File == (FILE *)NULL)
	{
		fprintf (stderr, "Cannot open source file %s: ", Source_Name);
		perror("");
		exit (1);
		/*NOTREACHED*/
	}

	/*
    **  Open the destination help file
    */

	Help_File = fopen (Help_Name, "wb");
	if (Help_File == (FILE *)NULL)
	{
		fprintf (stderr, "Cannot create help file %s: ", Help_Name);
		perror("");
		exit (1);
		/*NOTREACHED*/
	}

	/*
    **  Open the include file
    */

	strcpy (namebuf, Include_Name);
	strcat (namebuf, ".h");
	Include_File = fopen (namebuf, "w");
	if (Include_File == (FILE *)NULL)
	{
		fprintf (stderr, "Cannot create include file %s: ", namebuf);
		perror("");
		exit (1);
		/*NOTREACHED*/
	}
} /* Open_Files */


/*.
******************************************************************************
**
**  function -  Rewind_Source:
**
**  desc -      Seek to the beginning of the source file
**
**  entry -     globals:
**              Source_File
**
**  exit -      globals:
**              Line_Number
**
**  notes -
**
******************************************************************************
*/

void     Rewind_Source (void)

{
	rewind (Source_File);
	Line_Number = 0;
} /* Rewind_Source */


/*.
******************************************************************************
**
**  function -  Close_Files:
**
**  desc -      Close all files
**
**  entry -     globals:
**              Source_File
**              Help_File
**              Include_File
**
**  exit -
**
**  notes -     If an error occurs, a message is displayed and the program
**              terminates.
**
******************************************************************************
*/

static void     Close_Files (void)

{
	fclose (Source_File);
	fclose (Help_File);
	fclose (Include_File);
} /* Close_Files */


/*.
******************************************************************************
**
**  function -  main:
**
**  desc -      Program entry point and main block
**
**  entry -     argc, argv          standard command line args
**
**              See module header for command line syntax
**
**  exit -      See module header for program exit codes.
**
**  notes -
**
******************************************************************************
*/

int     main (int argc, char *argv[])

{
	Pgm_Name = argv[0];

	Title ();

	Sym_Init ();

	Parse_Command (argc, argv);

	Open_Files ();

	Pass1 ();

	Rewind_Source ();

	Pass2 ();

	Close_Files ();

	return (0);
} /* main */
